import threading
from django.utils.deprecation import MiddlewareMixin

# Variables de thread-local para almacenar la request y sus atributos
_thread_locals = threading.local()

def get_current_request():
    """
    Devuelve la request almacenada en el middleware.
    """
    return getattr(_thread_locals, 'request', None)

def set_current_request(request):
    """
        Modifica la request almacenada en el middleware.
    """
    return setattr(_thread_locals, 'request', request)

class AuditUserMiddleware(MiddlewareMixin):
    """
    Middleware combinado para capturar el usuario, la request, la URL, la IP y el dispositivo (user-agent).
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Obtener el usuario anterior desde la sesión
        previous_user = request.session.get('user_anterior', None)

        # Almacenar el usuario actual en thread-local antes de procesar la solicitud
        _thread_locals.previous_user = previous_user

        set_current_request(request)

        response = self.get_response(request)
        return response

    @staticmethod
    def get_current_user():
        """
        Devuelve el usuario actual de la solicitud.
        """
        request = get_current_request()
        user_id = getattr(request, 'user_id', None)
        if user_id:
            return user_id
        return None

    @staticmethod
    def get_previous_user():
        """
        Devuelve el usuario anterior almacenado en la sesión.
        """
        return getattr(_thread_locals, 'previous_user', None)

    @staticmethod
    def get_current_ip():
        """
        Devuelve la IP del cliente.
        """
        request = get_current_request()
        decoded_token = getattr(request, 'decoded_token', None)
        if decoded_token:
            datasession = decoded_token['dataUserSession']
            if datasession:
                ip = datasession['ip'] if datasession['ip'] else None
                return ip
            return None
        return None

    @staticmethod
    def get_user_agent():
        """
        Devuelve el User-Agent del cliente.
        """
        request = get_current_request()
        if request:
            return request.META.get('HTTP_USER_AGENT')
        return None

    @staticmethod
    def get_current_url():
        """
        Devuelve la URL completa de la solicitud actual.
        """
        request = get_current_request()
        if request:
            return request.path
        return None


import logging
from confluent_kafka.admin import AdminClient

class KafkaTopicManager:
    def __init__(self, bootstrap_servers):
        """Inicializa KafkaTopicManager con un AdminClient que se reutiliza."""
        if isinstance(bootstrap_servers, list):
            bootstrap_servers = ','.join(bootstrap_servers)
        self.admin_client = AdminClient({'bootstrap.servers': bootstrap_servers, 'log_level': 4})
        logging.info(f"AdminClient inicializado correctamente en {bootstrap_servers}")

    def ensure_topic_exists(self, topic):
        """Verifica si el topic existe."""
        topic_metadata = self.admin_client.list_topics(timeout=10)

        if topic not in topic_metadata.topics:
            raise Exception(f'ERROR EN KAFKA (TOPIC NO EXISTE): {topic}')


from django.apps import AppConfig, apps
from .helpers import *
from .logger import AuditLogger
from .kafka.main import KafkaProducer


class AuditLoggerConfig(AppConfig):
    name = 'audit_logger'

    def ready(self):
        bootstrap_servers = get_brokers()
        producer_name = get_project_name()
        KafkaProducer.initialize(bootstrap_servers, producer_name)

        self.register_all_audit()
    def register_all_audit(self):
        """Registra los modelos definidos por el usuario y todos los modelos para auditoría."""

        # Cargar los modelos definidos en audit_config_models.py
        models_config_by_user = load_audit_config_models()

        # Registrar auditoría para modelos definidos por el usuario
        if models_config_by_user:
            for model in models_config_by_user:
                if table_exists(model._meta.db_table):
                    AuditLogger.register_auditoria_config(model)
                else:
                    print(f"Skipping AUDITORIA {model.__name__}, table does not exist.")

        # Registrar auditoría para todos los modelos en la aplicación
        all_models = apps.get_models()
        for model in all_models:
            if table_exists(model._meta.db_table):
                AuditLogger.register_auditoria_logs(model)
            else:
                print(f"Skipping AUDITORIA {model.__name__}, table does not exist.")




from django.conf import settings
from django.db import connection
import importlib.util
import os

def get_project_name():
    """Obtiene el nombre del proyecto Django dinámicamente."""
    project_name = os.path.basename(settings.BASE_DIR)
    return project_name

def get_brokers():
    return getattr(settings, 'KAFKA_BROKERS_URLS', ['localhost:9092'])

def table_exists(table_name):
    """Verifica si la tabla existe en la base de datos."""
    with connection.cursor() as cursor:
        cursor.execute(f"SELECT 1 FROM information_schema.tables WHERE table_name = %s", [table_name])
        return cursor.fetchone() is not None

def get_settings_module_path():
    """Obtén la ruta donde se encuentra settings.py, usando el módulo de configuración."""
    from django.conf import settings

    # Obtener el módulo de configuración (settings.py) usando importlib
    settings_module = settings.SETTINGS_MODULE
    spec = importlib.util.find_spec(settings_module)

    if spec is None:
        raise RuntimeError(f"No se pudo encontrar el módulo {settings_module}")

    # Devolver la ruta del directorio que contiene settings.py
    return os.path.dirname(spec.origin)

def get_project_root(settings_module_path):
    """Obtén la ruta raíz del proyecto (un nivel por encima de settings_module_path)."""
    return os.path.dirname(settings_module_path)

def load_audit_config_models():
    """Carga los modelos definidos en audit_config_models.py."""
    try:
        # Obtener la ruta donde se encuentra settings.py
        settings_module_path = get_settings_module_path()

        # Primero buscar en la misma carpeta que settings.py
        audit_logger_path = os.path.join(settings_module_path, 'audit_config_models.py')

        # Si no se encuentra, probar un nivel por encima de settings_module_path
        if not os.path.exists(audit_logger_path):
            root_dir = get_project_root(settings_module_path)
            audit_logger_path = os.path.join(root_dir, 'audit_config_models.py')

        # Verificar si el archivo existe
        if not os.path.exists(audit_logger_path):
            raise FileNotFoundError(f"{audit_logger_path} not found.")

        # Cargar dinámicamente el archivo audit_config_models.py
        spec = importlib.util.spec_from_file_location("audit_config_models", audit_logger_path)
        audit_config_models = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(audit_config_models)

        # Devolver la lista de modelos definidos por el usuario
        return getattr(audit_config_models, 'CONFIG_MODELS', [])
    except (FileNotFoundError, AttributeError) as e:
        print(f"Error: {e}")
        return []


import logging
import asyncio

from .topic_manager import KafkaTopicManager
from .async_producer import KafkaAsyncProducer
from .local_logger import LocalAuditLogger

class KafkaProducer:
    _log_topic = 'audit_logs'
    _error_topic = 'audit_errors'
    _config_topic = 'audit_config'
    _producer = None
    _topic_manager = None
    _local_logger = None

    @classmethod
    def initialize(cls, bootstrap_servers, producer_name):
        """Inicializa los atributos de clase con los valores de configuración."""
        cls._producer = KafkaAsyncProducer(bootstrap_servers, producer_name)
        cls._topic_manager = KafkaTopicManager(bootstrap_servers)
        cls._local_logger = LocalAuditLogger()

    @classmethod
    async def _send_event(cls, topic, data):
        """Envía un evento a Kafka utilizando el productor asíncrono."""
        try:
            cls._topic_manager.ensure_topic_exists(topic)
            await cls._producer.send_event(topic, data)
        except Exception as kafka_error:
            logging.warning(f"Error al enviar mensaje a Kafka: {kafka_error}")
            await cls._local_logger.local_log(topic, data, str(kafka_error))

    @staticmethod
    def send_log_event(data):
        """Envía un evento de auditoría al topic de logs de manera asíncrona."""
        KafkaProducer._run_in_background(KafkaProducer._send_event, KafkaProducer._log_topic, data)

    @staticmethod
    def send_error_event(data):
        """Envía un evento de error al topic de errores de manera asíncrona."""
        KafkaProducer._run_in_background(KafkaProducer._send_event, KafkaProducer._error_topic, data)

    @staticmethod
    def send_config_event(data):
        """Envía un evento de configuración al topic de configuración de manera asíncrona."""
        KafkaProducer._run_in_background(KafkaProducer._send_event, KafkaProducer._config_topic, data)

    @staticmethod
    def _run_in_background(coro, *args):
        """Ejecuta una tarea en segundo plano, sin bloquear el flujo principal."""
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        if loop.is_running():
            asyncio.create_task(KafkaProducer._safe_run(coro, *args))
        else:
            loop.run_in_executor(None, lambda: asyncio.run(KafkaProducer._safe_run(coro, *args)))

    @staticmethod
    async def _safe_run(coro, *args):
        """Envuelve una tarea asíncrona en un manejador de errores para que no bloquee."""
        try:
            await coro(*args)
        except Exception as e:
            logging.error(f"Error en la ejecución asíncrona: {e}")

    @staticmethod
    async def close_producer():
        """Cierra el productor de Kafka."""
        if KafkaProducer._aioproducer is not None:
            await KafkaProducer._aioproducer.stop()
            KafkaProducer._aioproducer = None








import copy
import sys
import traceback
import inspect
from decimal import Decimal
from django.core.files import File
from datetime import date as DateClass, datetime as DateTimeClass, time
from django.db.models.signals import pre_save, post_save, post_delete
from django.forms.models import model_to_dict

from .middlewares import AuditUserMiddleware
from .kafka.main import KafkaProducer

class AuditLogger:
    previous_instance_state = {}
    registered_models_logs = set()  # Track models registered for log auditing
    registered_models_config = set()  # Track models registered for config auditing

    @staticmethod
    def register_auditoria_logs(model):
        """Registers a model for log auditing."""
        if model in AuditLogger.registered_models_config:
            print(f"Skipping {model.__name__} for logs as it's already registered for config.")
            return

        if model not in AuditLogger.registered_models_logs:
            AuditLogger._register_auditoria(model, 'logs')
            AuditLogger.registered_models_logs.add(model)
            print(f"Registered {model.__name__} for log auditing.")

    @staticmethod
    def register_auditoria_config(model):
        """Registers a model for configuration auditing."""
        # Unbind log auditing if it's already registered
        if model in AuditLogger.registered_models_logs:
            AuditLogger._unregister_auditoria_logs(model)
            print(f"Unregistered {model.__name__} from log auditing for config registration.")

        if model not in AuditLogger.registered_models_config:
            AuditLogger._register_auditoria(model, 'config')
            AuditLogger.registered_models_config.add(model)
            print(f"Registered {model.__name__} for configuration auditing.")

    @staticmethod
    def _unregister_auditoria_logs(model):
        """Unregisters a model from log auditing."""
        pre_save.disconnect(AuditLogger._audit_pre_save, sender=model)
        post_save.disconnect(sender=model)
        post_delete.disconnect(sender=model)
        AuditLogger.registered_models_logs.discard(model)
        print(f"Model {model.__name__} unregistered from log auditing.")

    @staticmethod
    def register_auditoria_errors(exception):
        """Captures system errors for auditing."""
        AuditLogger._audit_error(exception)

    @staticmethod
    def _register_auditoria(model, event_type):
        """General method to register models for auditing."""
        pre_save.connect(AuditLogger._audit_pre_save, sender=model)
        post_save.connect(
            lambda sender, instance, created, **kwargs: AuditLogger._audit_post_save(sender, instance, created, event_type),
            sender=model, weak=False
        )
        post_delete.connect(
            lambda sender, instance, **kwargs: AuditLogger._audit_post_delete(sender, instance, event_type),
            sender=model, weak=False
        )

    @staticmethod
    def _audit_pre_save(sender, instance, **kwargs):
        """Captures the previous state before changes are saved."""
        if instance.pk:
            try:
                previous_instance = sender.objects.get(pk=instance.pk)
                AuditLogger.previous_instance_state[instance.pk] = copy.deepcopy(previous_instance)
            except sender.DoesNotExist:
                pass

    @staticmethod
    def _audit_post_save(sender, instance, created, event_type, **kwargs):
        """Audits actions after a model is saved."""
        current_user, previous_user, context_info = AuditLogger._get_user_context(sender, instance)
        previous_instance = AuditLogger.previous_instance_state.get(instance.pk)

        previous_state = AuditLogger._serialize_instance(previous_instance) if previous_instance else None
        current_state = AuditLogger._serialize_instance(instance)
        action = 'created' if created else 'updated'

        # Handle logical deletion
        if not created and previous_state and current_state.get('es_activo') == False and previous_state.get('es_activo') == True:
            action = 'logic_deleted'

        audit_data = AuditLogger._build_audit_data(
            instance, sender, action, current_user, previous_user, previous_state, current_state, **context_info
        )

        AuditLogger._send_event(event_type, audit_data)
        AuditLogger.previous_instance_state.pop(instance.pk, None)

    @staticmethod
    def _audit_post_delete(sender, instance, event_type, **kwargs):
        """Audits actions after a model is deleted."""
        current_user, previous_user, context_info = AuditLogger._get_user_context(sender, instance)
        previous_state = AuditLogger._serialize_instance(instance) if instance else None

        audit_data = AuditLogger._build_audit_data(
            instance, sender, 'deleted', current_user, previous_user, previous_state, None, **context_info
        )

        AuditLogger._send_event(event_type, audit_data)

    @staticmethod
    def _audit_error(exception):
        """Handles error auditing."""
        exc_type, _, exc_tb = sys.exc_info()
        error_type = exc_type.__name__
        file_name = exc_tb.tb_frame.f_code.co_filename
        line_number = exc_tb.tb_lineno

        current_user, _, context_info = AuditLogger._get_user_context()

        audit_data = {
            "error_type": error_type,
            "error_message": str(exception),
            "stack_trace": traceback.format_exc(),
            "file_name": file_name,
            "line_number": line_number,
            "timestamp": str(DateTimeClass.now()),
            "current_user": current_user or 'Anonymous',
            **context_info
        }

        KafkaProducer.send_error_event(audit_data)

    @staticmethod
    def _build_audit_data(instance, sender, action, current_user, previous_user, previous_state, current_state, **context):
        """Builds the audit data dictionary."""
        frame = inspect.stack()[2]
        event_file = frame.filename
        event_line = frame.lineno

        return {
            "model": sender.__name__,
            "instance_id": getattr(instance, 'id', None),
            "description": getattr(instance, 'auditoria_descripcion', None),
            "action": action,
            "current_user": current_user or None,
            "previous_user": previous_user or None,
            "module": sender.__module__,
            "previous_state": previous_state,
            "current_state": current_state,
            "timestamp": str(DateTimeClass.now()),
            "event_file": event_file,
            "event_line": event_line,
            **context
        }

    @staticmethod
    def _get_user_context(sender=None, instance=None):
        """Gets the current user, previous user, and other contextual info."""
        current_user = AuditUserMiddleware.get_current_user()
        previous_user = AuditUserMiddleware.get_previous_user()
        context = {
            "url": AuditUserMiddleware.get_current_url(),
            "ip_address": AuditUserMiddleware.get_current_ip(),
            "user_agent": AuditUserMiddleware.get_user_agent(),
        }

        return current_user, previous_user, context

    @staticmethod
    def _send_event(event_type, audit_data):
        """Sends the appropriate event to Kafka."""
        if event_type == 'logs':
            KafkaProducer.send_log_event(audit_data)
        elif event_type == 'config':
            KafkaProducer.send_config_event(audit_data)

    @staticmethod
    def _serialize_instance(instance):
        """Converts a model instance to a serializable dictionary, including ManyToMany fields."""
        # Convertir los campos regulares con model_to_dict
        data = model_to_dict(instance, exclude=[field.name for field in instance._meta.many_to_many])

        # Serializar campos ManyToMany
        for field in instance._meta.many_to_many:
            value = getattr(instance, field.name)
            data[field.name] = list(value.values_list('id', flat=True))  # Solo los IDs de los objetos relacionados

        # Procesar otros tipos de campos
        for field in instance._meta.fields:
            value = getattr(instance, field.name)

            if isinstance(value, File):
                data[field.name] = value.url if value else None
            elif isinstance(value, Decimal):
                data[field.name] = float(value)
            elif isinstance(value, (DateClass, DateTimeClass)):
                data[field.name] = value.isoformat()
            elif isinstance(value, time):
                data[field.name] = value.strftime('%H:%M:%S')

        return data


import json
import logging
from django.db import connection
from asgiref.sync import sync_to_async

class LocalAuditLogger:
    """Clase que maneja el registro de errores de Kafka en PostgreSQL sin migraciones."""

    def __init__(self):
        """Inicializa el auditor local y crea la tabla si no existe."""
        self._create_table_if_not_exists()

    def _create_table_if_not_exists(self):
        """Crea la tabla kafka_local_logs si no existe en la base de datos."""
        create_table_query = '''
        CREATE TABLE IF NOT EXISTS kafka_local_logs (
            id SERIAL PRIMARY KEY,
            error TEXT NOT NULL,
            registro JSONB NOT NULL,
            topic VARCHAR(255) NOT NULL,
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
        );
        '''
        try:
            with connection.cursor() as cursor:
                cursor.execute(create_table_query)
                cursor.execute("SELECT to_regclass('public.kafka_local_logs')")
                table_exists = cursor.fetchone()
                if table_exists[0]:
                    logging.info("La tabla kafka_local_logs ya existía en la base de datos.")
                else:
                    logging.info("La tabla local kafka_local_logs fue creada exitosamente.")
        except Exception as e:
            logging.error(f"Error al crear/verificar la tabla kafka_local_logs: {e}")

    @sync_to_async
    def local_log(self, topic, data, error_message):
        """Guarda el registro fallido en la base de datos PostgreSQL."""
        insert_query = '''
        INSERT INTO kafka_local_logs (error, registro, topic)
        VALUES (%s, %s, %s);
        '''
        with connection.cursor() as cursor:
            try:
                cursor.execute(insert_query, (error_message, json.dumps(data), topic))
                logging.info(f"Registro guardado localmente en PostgreSQL motivo: {error_message}")
            except Exception as db_error:
                logging.error(f"Error al guardar el mensaje en PostgreSQL: {db_error}")


from setuptools import setup, find_packages

setup(
    name='unemi-audi-kafka',
    version='1.1.10',
    packages=find_packages(),
    install_requires=[
        'Django>=3.1',
        'confluent_kafka>=1.7.0',
        'aiokafka>=0.11.0',
    ],
    description='Una aplicación reutilizable de Django para auditar modelos automáticamente con integración con Kafka.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',  # Specify that the description is in Markdown
    author='Marcos Daniel',
    author_email='mteranc@unemi.edu.ec',
    url='https://github.com/marcosdaniel2002/django-audit-kafkalogger',
    classifiers=[
        'Framework :: Django',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    license='MIT',
    python_requires='>=3.6',
)



from aiokafka import AIOKafkaProducer
import logging
import json

class KafkaAsyncProducer:
    """Clase que maneja el productor asíncrono de Kafka."""
    def __init__(self, bootstrap_servers, producer_name):
        if isinstance(bootstrap_servers, str):
            bootstrap_servers = [bootstrap_servers]
        self._bootstrap_servers = bootstrap_servers
        self._producer_name = producer_name
        self._aioproducer = None

    async def _get_aioproducer(self):
        """Crea y configura el productor asíncrono de Kafka si no está inicializado."""
        if self._aioproducer is None:
            try:
                self._aioproducer = AIOKafkaProducer(
                    bootstrap_servers=self._bootstrap_servers,
                    client_id=self._producer_name
                )
                await self._aioproducer.start()
            except Exception as e:
                logging.error(f"Error al conectar con Kafka: {e}")
                raise
        return self._aioproducer

    async def send_event(self, topic, data):
        """Envía un evento a Kafka de manera asíncrona."""
        producer = await self._get_aioproducer()

        if isinstance(data, dict):
            data['producer'] = self._producer_name
            data = json.dumps(data, ensure_ascii=False).encode('utf-8')  # Convierte el dict a JSON
        else:
            logging.error("El dato proporcionado no es un diccionario y no se puede serializar.")
            return

        try:
            await producer.send_and_wait(topic, data)
            logging.info(f"Mensaje enviado correctamente a {topic}")
        except Exception as kafka_error:
            logging.error(f"Error al enviar mensaje a Kafka: {kafka_error}")
            raise

    async def close(self):
        """Cierra el productor de Kafka."""
        if self._aioproducer is not None:
            await self._aioproducer.stop()
            self._aioproducer = None